#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Interfaces for nti.transactions.

"""

from __future__ import print_function, absolute_import, division

from zope.interface import Interface
from zope.interface import Attribute
from zope.interface import implementer

from transaction.interfaces import TransactionError
from transaction.interfaces import ITransaction

# pylint:disable=no-method-argument,inherit-non-class
# Sigh. The Python 2 version of pylint raises this.
# pylint:disable=too-many-ancestors

class IExtendedTransaction(ITransaction):
    """Extensions to the transaction api."""

    def nti_commit():
        """
        Like ``commit``, but produces a :obj:`perfmetrics.Metric` ``transaction.commit``
        metric.
        """

    def nti_abort():
        """
        Like ``abort``, but produces a :obj:`perfmetrics.Metric`
        ``transaction.abort`` metric.
        """

class CommitFailedError(TransactionError):
    """
    Committing the active transaction failed for an unknown
    and unexpected reason.

    This is raised instead of raising very generic system exceptions such as
    TypeError.
    """

class AbortFailedError(TransactionError):
    """
    Aborting the active transaction failed for an unknown and unexpected
    reason.

    This is raised instead of raising very generic system exceptions
    such as ValueError and AttributeError.
    """

class TransactionLifecycleError(TransactionError):
    """
    Raised when an application commits or aborts a transaction
    while the transaction controller believes it is in control.

    *Applications must not raise this exception.*

    This may have happened many times; we cannot detect that.

    This is a programming error.
    """

class ForeignTransactionError(TransactionLifecycleError):
    """
    Raised when a transaction manager has its transaction changed
    while a controlling transaction loop believes it is in control.

    The handler first aborted or committed the transaction, and then
    began a new one. Possibly many times.

    A kind of `TransactionLifecycleError`. *Applications must not
    raise this exception.*

    This is a programming error.
    """

class ILoopInvocation(Interface):
    """
    Description of why a loop was invoked.
    """
    handler = Attribute("The handler to run.")
    loop = Attribute("The loop doing the running.")
    args = Attribute("The arguments passed to the handler")
    kwargs = Attribute("The keyword arguments passed to the handler.")

class ILoopEvent(Interface):
    """
    Base class for event loop events.

    The *handler_args* and *handler_kwargs* should not
    be mutated in place or assigned to. These are a convenience
    for accessing the attributes of the *invocation* attribute.

    .. versionchanged:: 4.2.0
       Add the *handler_args* and *handler_kwargs*.
    """
    invocation = Attribute("An ILoopInvocation.")
    handler_args = Attribute("The positional arguments for the handler, or ()")
    handler_kwargs = Attribute("The keyword arguments for the handler, or {}")

class IAfterTransactionBegan(ILoopEvent):
    """
    A new transaction has begun.
    """

    tx = Attribute("The transaction.")

class IWillAttemptTransaction(ILoopEvent):
    """
    Base class for attempt events.
    """
    tx = Attribute("The transaction.")
    attempt_number = Attribute("The number of the attempt. Starts at 0.")

# We get this in Python 2
# pylint:disable=too-many-ancestors

class IWillFirstAttempt(IWillAttemptTransaction):
    """
    The first attempt.
    """

class IWillRetryAttempt(IWillAttemptTransaction):
    """
    A retry attempt.
    """

class IWillLastAttempt(IWillRetryAttempt):
    """
    The final retry.
    """

class IWillSleepBetweenAttempts(ILoopEvent):
    """
    Will sleep between attempts.

    If the ``sleep_time`` attribute is modified,
    that will be the time slept.
    """
    sleep_time = Attribute("The time to sleep.")


class IWillFirstAttemptWithRequest(IWillFirstAttempt):
    """
    The first attempt, but with a request object.

    .. versionadded:: 4.2.0
    """
    request = Attribute("The request object that will be used.")


class IWillRetryAttemptWithRequest(IWillRetryAttempt):
    """
    A retry attempt with a request object.

    .. versionadded:: 4.2.0
    """
    request = Attribute("The request object that will be used.")


class IWillLastAttemptWithRequest(IWillLastAttempt):
    """
    The final retry with a request object.

    .. versionadded:: 4.2.0
    """
    request = Attribute("The request object that will be used.")

@implementer(ILoopInvocation)
class LoopInvocation(object):
    __slots__ = ('loop', 'handler', 'args', 'kwargs')

    def __init__(self, loop, handler, args, kwargs):
        self.loop = loop
        self.handler = handler
        self.args = args
        self.kwargs = kwargs


@implementer(ILoopEvent)
class LoopEvent(object):
    __slots__ = ('invocation',)
    def __init__(self, invocation):
        self.invocation = invocation

    @property
    def handler_args(self):
        return self.invocation.args

    @property
    def handler_kwargs(self):
        return self.invocation.kwargs

@implementer(IAfterTransactionBegan)
class AfterTransactionBegan(LoopEvent):
    __slots__ = ('tx',)
    def __init__(self, invocation, tx):
        LoopEvent.__init__(self, invocation)
        self.tx = tx

@implementer(IWillAttemptTransaction)
class WillAttemptTransaction(LoopEvent):
    __slots__ = ('tx', 'attempt_number')
    def __init__(self, invocation, tx, attempt_number):
        LoopEvent.__init__(self, invocation)
        self.tx = tx
        self.attempt_number = attempt_number

@implementer(IWillFirstAttempt)
class WillFirstAttempt(WillAttemptTransaction):
    __slots__ = ()

@implementer(IWillRetryAttempt)
class WillRetryAttempt(WillAttemptTransaction):
    __slots__ = ()

@implementer(IWillLastAttempt)
class WillLastAttempt(WillRetryAttempt):
    __slots__ = ()

@implementer(IWillSleepBetweenAttempts)
class WillSleepBetweenAttempts(LoopEvent):
    __slots__ = ('sleep_time',)

    def __init__(self, invocation, sleep_time):
        LoopEvent.__init__(self, invocation)
        self.sleep_time = sleep_time

class _RequestMixin(object):
    @property
    def request(self):
        return self.handler_args[0]

@implementer(IWillFirstAttemptWithRequest)
class WillFirstAttemptWithRequest(WillFirstAttempt, _RequestMixin):
    __slots__ = ()

@implementer(IWillRetryAttemptWithRequest)
class WillRetryAttemptWithRequest(WillRetryAttempt, _RequestMixin):
    __slots__ = ()

@implementer(IWillLastAttemptWithRequest)
class WillLastAttemptWithRequest(WillLastAttempt, _RequestMixin):
    __slots__ = ()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import unittest

from hamcrest import assert_that
from hamcrest import contains_exactly as contains
from hamcrest import calling
from hamcrest import raises
from hamcrest import is_

from ..manager import do
from ..manager import do_near_end

import transaction


from nti.testing.base import AbstractTestBase


class TestDataManagerFunctions(unittest.TestCase):

    def test_data_manager_sorting(self):
        results = []
        def test_call(x):
            results.append(x)

        # The managers will execute in order added (since identical),
        # except for the one that requests to go last.
        do(call=test_call, args=(0,))
        do(call=test_call, args=(1,))
        do_near_end(call=test_call, args=(10,))
        do(call=test_call, args=(2,))
        transaction.commit()
        assert_that(results, contains(0, 1, 2, 10))


class TestObjectDataManager(AbstractTestBase):

    def test_vote(self):
        class Exc(Exception):
            pass
        def vote():
            raise Exc()

        odm = do(call=lambda: 1, vote=vote)
        assert_that(calling(odm.tpc_vote).with_args(None), raises(Exc))

    def test_callable_name(self):
        class X(object):
            def thing(self):
                """Nothing"""

        x = X()
        odm = do(target=x, method_name='thing')
        assert_that(odm.callable, is_(x.thing))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A Pyramid tween that begins and ends transactions around its handler
using the :class:`nti.transactions.loop.TransactionLoop`.

This very similar to earlier versions of :mod:`pyramid_tm`, but with
the following substantial differences:

    - The transaction is rolled back if the request is deemed to be
      side-effect free (this has intimate knowledge of the paths that
      do not follow the HTTP rules for a GET being side-effect free;
      however, if you are a GET request and you violate the rules by
      having side-effects, you can set the environment key
      ``nti.request_had_transaction_side_effects`` to ``True``)

    - Logging is added to account for the time spent in aborts and
      commits.

    - Later versions of :mod:`pyramid_tm` were split into two parts,
      with :mod:`pyramid_retry` being used to handle retry using an
      "execution policy", which was new in Pyramid 1.9. This library
      is compatible with all versions of Pyramid from 1.2 onward.

Install this tween using the ``add_tween`` method::

    pyramid_config.add_tween(
        'nti.transactions.pyramid_tween.transaction_tween_factory',
        under=pyramid.tweens.EXCVIEW)

You may install it under or over the exception view, depending on whether you
need the transaction to be active in exception views.

If you have a tween that manages a ZODB connection, it should be installed
*above* this tween. That's because ZODB connections cannot be closed while
joined to a transaction; the transaction must be committed or aborted first.
"""

import logging


from ._httpexceptions import HTTPBadRequest
from ._httpexceptions import HTTPException
from ._loglevels import TRACE
from .loop import TransactionLoop
from .interfaces import WillFirstAttemptWithRequest
from .interfaces import WillRetryAttemptWithRequest
from .interfaces import WillLastAttemptWithRequest

logger = logging.getLogger(__name__)

__all__ = [
    'commit_veto',
    'is_side_effect_free',
    'is_error_retryable',
    'is_last_attempt',
    'TransactionTween',
    'transaction_tween_factory',
]

def commit_veto(request, response):
    """
    When used as a commit veto, the logic in this function will cause
    the transaction to be aborted if:

        - An ``X-Tm`` response header with the value ``'abort'`` (or
          any value other than ``'commit'``) exists. A value of
          ``'commit'`` overrides any later match in this list.

        - The response status code starts with ``'4'`` or ``'5'``.

        - The request environment has a true value for
          ``'nti.commit_veto'``

    Otherwise the transaction will be allowed to commit.
    """
    xtm = response.headers.get('x-tm')
    if xtm is not None:
        return xtm != 'commit'
    if response.status.startswith(('4', '5')):
        return True
    return request.environ.get('nti.commit_veto')

def is_side_effect_free(request):
    """
    Is the request side-effect free? If the answer is yes, we should
    be able to quietly abort the transaction and avoid taking out any
    locks in the DBs.

    A request is considered to be free of side effects if:

        - It is a GET or HEAD request; AND

        - The URL DOES NOT match a configured exception list.

    In this version, the configured exception list is not actually
    configurable. It is hardcoded to handle the case of socket.io
    polling (which does have side effects on GET requests), while
    still supporting serving the static resources of socket.io.
    """
    if request.method in {'GET', 'HEAD'}:
        # GET/HEAD requests must NEVER have side effects.
        if 'socket.io' in request.url:
            # (Unfortunately, socket.io polling does)
            # However, the static resources don't.

            # TODO: This needs to be configurable *and* fast.
            #
            # A hardcoded nested conditional using `in` takes 224ns when the
            # answer is yes, and 218ns when it's no but still contains
            # socket.io, and 178ns when it's not even close --- we
            # expect that last case to be the common case. Changing the
            # first 'in' to be startswith('/socket.io') counter-intuitively slows
            # down the common case to be 323ns.
            #
            # A hardcoded regex match takes 438ns when the answer is yes,
            # 455ns when its no but still starts with socket.io, and
            # 390ns when it's not close. So twice as bad in the common case.
            # (All timings from CPython 2.7; revisit on PyPy and Python 3)
            return 'static' in request.url
        return True
    # Every non-get probably has side effects
    return False


def is_last_attempt(request):
    """
    Return `True` if the request is on its last attempt, meaning that
    the `TransactionTween` will not invoke the handler again, regardless
    of what happens during this attempt.
    """
    environ = request.environ
    attempt = environ.get('retry.attempt')
    attempts = environ.get("retry.attempts")

    return True if attempt is None or attempts is None else attempt + 1 == attempts


def is_error_retryable(request, exc):
    """
    Return `True` if the exception is one that can be retried.

    This will return `False` if the request is on its last attempt.
    """
    if is_last_attempt(request) or 'nti.transaction_tween_retryable' not in request.environ:
        return False

    return request.environ['nti.transaction_tween_retryable'](exc)


class TransactionTween(TransactionLoop):
    """
    A Pyramid tween for retrying execution of a request.

    When the request body is executing, the functions :func:`is_last_attempt`
    and :func:`is_error_retryable` can be used to influence handler behaviour.

    .. versionchanged:: 4.2.0
       For convenience, now emits :class:`nti.transactions.interfaces.IWillFirstAttemptWithRequest`,
       and the other ``WithRequest`` events, instead of events without the
       request object.
    """

    EVT_WILL_FIRST_ATTEMPT = WillFirstAttemptWithRequest
    EVT_WILL_RETRY_ATTEMPT = WillRetryAttemptWithRequest
    EVT_WILL_LAST_ATTEMPT = WillLastAttemptWithRequest

    def prep_for_retry(self, attempts_remaining, tx, request): # pylint:disable=arguments-differ
        """
        Prepares the request for possible retries.

        Buffers the body if needed using
        :meth:`pyramid.request.Request.make_body_seekable`.

        The first time this is called for a given request, if the method is
        expected to have a body and the body appears to be JSON, but
        the content type specifies a browser form submission, the
        content type is changed to be ``application/json``. This is a
        simple fix for broken clients that forget to set the HTTP
        content type. This may be removed in the future.

        .. caution::
            This doesn't do anything with the WSGI
            environment; changes to that persist
            between retries.

        .. versionchanged:: 4.2.0
           Now, transient attributes of the request object (those whose
           name begins with ``_v_``) are deleted by this function before
           it returns for the second and susquent attempts.
           Previously, there were left with their existing values.
           This is intended to make request-specific caching easier,
           following a similar model as :mod:`persistent` and allowing for
           common patterns such as :class:`zope.cachedescriptors.property.Lazy`.
        """
        # make_body_seekable will copy wsgi.input if necessary,
        # otherwise it will rewind the copy to position zero
        try:
            request.make_body_seekable()
        except IOError as e:
            # almost always " unexpected end of file reading request";
            # (though it could also be a tempfile issue if we spool to
            # disk?) at any rate,
            # this is non-recoverable
            logger.log(TRACE, "Failed to make request body seekable",
                       exc_info=True)
            # TODO: Should we do anything with the request.response? Set an error
            # code? It won't make it anywhere...

            # However, it is critical that we return a valid Response
            # object, even if it is an exception response, so that
            # Pyramid doesn't blow up

            raise self.AbortException(HTTPBadRequest(str(e)),
                                      "IOError on reading body")

        attempts = self.attempts
        # Starting at 0 for compatibility with pyramid_retry.
        attempt_number = attempts - attempts_remaining - 1
        request.environ['retry.attempt'] = attempt_number
        request.environ['retry.attempts'] = attempts
        request.environ['nti.transaction_tween_retryable'] = lambda ex: self._retryable(
            tx, (None, ex, None)
        )

        # WebTest, browsers, and many of our integration tests by
        # default sets a content type of
        # 'application/x-www-form-urlencoded' If you happen to access
        # request.POST, though, (like locale negotiation does, or
        # certain template operations do) the underlying WebOb will
        # notice the content-type and attempt to decode the body based
        # on that. This leads to a badly corrupted body (if it was
        # JSON) and mysterious failures; this has been seen in the
        # real world. An internal implementation change (accessing
        # POST) suddenly meant that we couldn't read their body.
        # Unfortunately, the mangling is not fully reversible, since
        # it wasn't encoded in the first place.

        # We attempt to fix that here. (This is the best place because
        # we are now sure the body is seekable.)
        # pylint:disable-next=compare-to-zero
        if attempt_number == 0 \
           and request.method in {'POST', 'PUT'} \
           and request.content_type == 'application/x-www-form-urlencoded':
            # This needs tested.
            body = request.body
            # Python 3 treats bytes different than strings. Iteration
            # and indexing don't iterate over one-byte strings, they return
            # *integers*. 123 and 91 are the integers for { and [
            if body and body[0] in {b'{', b'[', 123, 91}:
                # encoded data will never start with these values, they would be
                # escaped. so this must be meant to be JSON
                request.content_type = 'application/json'
        if attempt_number > 0:
            # We are making a retry. We need to reset transient state.
            for attr_name in list(attr_name
                                  for attr_name
                                  in vars(request)
                                  if attr_name.startswith('_v_')):
                delattr(request, attr_name)

    def should_abort_due_to_no_side_effects(self, request): # pylint:disable=arguments-differ
        """
        Tests with :func:`is_side_effect_free`.

        If the request's ``environ`` has a true value for the key
        ``'nti.request_had_transaction_side_effects'``, this method will return
        false.
        """
        return  is_side_effect_free(request) and \
                not request.environ.get('nti.request_had_transaction_side_effects')

    def should_veto_commit(self, response, request): # pylint:disable=arguments-differ
        """
        Tests with :func:`commit_veto`.
        """
        return commit_veto(request, response)

    def describe_transaction(self, request): # pylint:disable=arguments-differ
        return request.path_info

    def run_handler(self, request): # pylint:disable=arguments-differ
        try:
            return TransactionLoop.run_handler(self, request)  # Not super() for speed
        except HTTPException as e:
            # Pyramid catches these and treats them as a response. We
            # MUST catch them as well and let the normal transaction
            # commit/doom/abort rules take over--if we don't catch
            # them, everything appears to work, but the exception
            # causes the transaction to be aborted, even though the
            # client gets a response.
            #
            # The problem with simply catching exceptions and returning
            # them as responses is that it bypasses pyramid's notion
            # of "exception views". At this writing, we are only
            # using those to turn 403 into 401 when needed, but it
            # can also be used for other things (such as redirecting what
            # would otherwise be a 404).
            # So we wrap up __call__ and also check for HTTPException there
            # and raise it safely after transaction handling is done.
            # Of course, this is only needed if the exception was actually
            # raised, not deliberately returned (commonly HTTPFound and the like
            # are returned)...raising those could have unintended consequences

            request._v_nti_raised_exception = True # pylint:disable=protected-access
            return e

    def __call__(self, request): # pylint:disable=arguments-differ
        result = TransactionLoop.__call__(self, request)  # not super() for speed
        if  isinstance(result, HTTPException) and \
            getattr(request, '_v_nti_raised_exception', False):
            raise result
        return result

def transaction_tween_factory(handler, registry):
    """
    The factory to create the tween.

    The *registry* argument from Pyramid is used to access the
    Pyramid "deployment settings" to configure the behaviour
    of the transaction loop. The following settings are used.

    retry.attempts
        Int. The number of retry attempts. See :attr:`.TransactionLoop.attempts`
        for the default.
    retry.long_commit_duration
        Float. The number of seconds after which a commit will be considered
        "too long" and a warning logging message issued. See
        :attr:`.TransactionLoop.long_commit_duration` for the default.
    retry.sleep_ms
        Int. The base number of **milliseconds** to sleep between retry attempts.
        See :attr:`.TransactionLoop.sleep` for more documentation.

    See :class:`TransactionTween`
    """
    settings = registry.settings
    def setting(name, converter):
        v = settings.get(name, None)
        if v is not None:
            v = converter(v)
        return v

    attempts = setting('retry.attempts', int)
    retries = attempts - 1 if attempts else None

    sleep_ms = setting('retry.sleep_ms', int)
    sleep = sleep_ms / 1000 if sleep_ms else None

    long_commit_duration = setting('retry.long_commit_duration', float)


    tween = TransactionTween(handler,
                             retries=retries,
                             sleep=sleep,
                             long_commit_duration=long_commit_duration)
    logger.info("Created tween %s", tween)
    return tween



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Support for transactionally working with queues.
"""

import transaction

from queue import Full as QFull


from .manager import ObjectDataManager

__all__ = [
    'put_nowait',
]


class _QueuePutDataManager(ObjectDataManager):
    """
    A data manager that checks if the queue is full before putting.
    Overrides :meth:`tpc_vote` for efficiency.
    """

    def __init__(self, queue, method, args=()):
        super().__init__(target=queue, call=method, args=args)
        # NOTE: See the `sortKey` method. The use of the queue as the target
        # is critical to ensure that the FIFO property holds when multiple objects
        # are added to a queue during a transaction

    def tpc_vote(self, tx):
        if self.target.full():
            # TODO: Should this be a transient exception?
            # So retry logic kicks in?
            raise QFull()

def put_nowait(queue, obj):
    """
    Transactionally puts `obj` in `queue`. The `obj` will only be visible
    in the queue after the current transaction successfully commits.
    If the queue cannot accept the object because it is full, the transaction
    will be aborted.

    See :class:`gevent.queue.Queue` and :class:`Queue.Full` and :mod:`gevent.queue`.
    """
    transaction.get().join(
        _QueuePutDataManager(queue,
                             queue.put_nowait,
                             args=(obj,)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Support for working with the :mod:`transaction` module.

This module imports the :mod:`dm.transaction.aborthook` module and
directly provides the :func:`add_abort_hooks` function; you should
call this if you need such functionality.

"""

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

from logging import DEBUG
from logging import WARNING
from logging import ERROR
from logging import getLogger

import sys
import time
import random

import six

from perfmetrics import statsd_client as _statsd_client

from zope.exceptions.exceptionformatter import format_exception
from zope.exceptions.exceptionformatter import print_exception
from zope.cachedescriptors.property import Lazy
from zope.event import notify

from .interfaces import CommitFailedError
from .interfaces import AbortFailedError
from .interfaces import ForeignTransactionError
from .interfaces import TransactionLifecycleError

from .interfaces import LoopInvocation
from .interfaces import AfterTransactionBegan
from .interfaces import WillFirstAttempt
from .interfaces import WillRetryAttempt
from .interfaces import WillLastAttempt
from .interfaces import WillSleepBetweenAttempts

import transaction

from transaction.interfaces import AlreadyInTransaction
from transaction.interfaces import NoTransaction

from nti.transactions import DEFAULT_LONG_RUNNING_COMMIT_IN_SECS

__all__ = [
    'TransactionLoop',
]

logger = getLogger(__name__)

def _do_commit(tx, long_commit_duration,
               retries, sleep_time,
               _logger=logger,
               _DEBUG=DEBUG,
               _WARNING=WARNING,
               _perf_counter=getattr(time, 'perf_counter', time.time)):
    exc_info = sys.exc_info()
    try:
        begin = _perf_counter()
        tx.nti_commit()
        level = _DEBUG
        duration = _perf_counter() - begin
        if duration > long_commit_duration:
            level = _WARNING
        if _logger.isEnabledFor(level):
            _logger.log(
                level,
                "Committed transaction description=%r, duration=%s, retries=%s, sleep_time=%s",
                tx.description, duration, retries, sleep_time
            )
    except TypeError:
        # Translate this into something meaningful
        exc_info = sys.exc_info()
        six.reraise(CommitFailedError, CommitFailedError(exc_info[1]), exc_info[2])
    except (AssertionError, ValueError):
        # We've seen this when we are recalled during retry handling. The higher level
        # is in the process of throwing a different exception and the transaction is
        # already toast, so this commit would never work, but we haven't lost anything;
        # The sad part is that this assertion error overrides the stack trace for what's currently
        # in progress

        # TODO: Prior to transaction 1.4.0, this was only an
        # AssertionError. 1.4 makes it a ValueError, which is hard to
        # distinguish and might fail retries?
        logger.exception("Failing to commit; should already be an exception in progress")
        if exc_info and exc_info[0]:
            six.reraise(*exc_info)

        raise
    finally:
        del exc_info

class TransactionLoop(object):
    """
    Similar to the transaction attempts mechanism, but less error
    prone and with added logging and hooks.

    This object is callable (passing any arguments along to its
    handler) and runs its handler in the transaction loop.

    The handler code may doom the transaction, but they must not
    attempt to commit or abort it. A doomed transaction, or one whose
    commit is vetoed by :meth:`should_abort_due_to_no_side_effects` or
    :meth:`should_veto_commit` is never retried.

    Aborting or committing the transaction will set these :mod:`perfmetrics`
    timers:

    transaction.commit
        Time taken to commit the transaction. Sampled.
    transaction.abort
        Time taken to abort the transaction. Sampled.

    Running the loop will increment these :mod:`perfmetrics` counters
    (new in 3.1):

    transaction.retry
        The number of retries required in any given loop. Note that
        if the handler raises non-retryable exceptions, this number will
        be inflated.
    transaction.side_effect_free
        How many side-effect free transactions there have been.
    transaction.side_effect_free_violation
        How many side-effect free transactions actually had resource
        managers joined to the transaction, and so potentially aborted
        work that wanted to be committed. (3.1.1)
    transaction.vetoed
        How many transactions were vetoed.
    transaction.doomed
        The number of doomed transactions.
    transaction.successful
        The number of transactions that successfully returned.
    transaction.failed
        The number of transactions that did not successfully return.

    .. important::

       Instances of this class must be thread safe, and running the loop should
       not mutate the state of this object.

    .. versionchanged:: 3.0

       When this object is called, the thread-local default or global
       :class:`transaction.TransactionManager` is placed into explicit
       mode (if it wasn't already). The handler callable is forbidden
       from beginning, aborting or committing the transaction. Explicit
       transactions can be faster in ZODB, and are easier to reason
       about.

       If the application begins, commits or aborts the transaction, it
       can expect this object to raise
       :exc:`transaction.interfaces.NoTransaction`,
       :exc:`transaction.interfaces.AlreadyInTransaction` or
       :exc:`nti.transactions.interfaces.TransactionLifecycleError`.

    .. versionchanged:: 3.1

       zope.event is used to publish events after each transaction begins,
       before a transaction is retried or the first attempt is made,
       and before we sleep between retries. See
       :class:`nti.transaction.interfaces.IAfterTransactionBegan`,
       :class:`nti.transaction.interfaces.IWillFirstAttempt`,
       :class:`nti.transaction.interfaces.IWillRetryAttempt`,
       :class:`nti.transaction.interfaces.IWillSleepBetweenAttempt`.
    """

    class AbortAndReturn(Exception):
        """
        This private exception is raised here to cause us to break out
        of our loop, abort the transaction, and return the result.

        .. versionchanged:: 3.0

           Previously this was called ``AbortException``. Until 4.0,
           that name remains available as a deprecated alias.
        """
        def __init__(self, response, reason):
            Exception.__init__(self)
            self.response = response
            self.reason = reason

    #: Deprecated alias for `AbortAndReturn`
    AbortException = AbortAndReturn

    #: If not None, this is a number that specifies the base amount of time
    #: (in seconds) to wait after a failed transaction attempt before retrying. Each
    #: retry attempt will pick a delay following the binary exponential
    #: backoff algorithm: ``sleep * (random number between 0 and 2^retry-1)``.
    #: (A simple increasing multiplier might work well if there is only one other
    #: transaction that we conflict with, but in cases of multiple conflicts
    #: or even new conflicts, the random backoff should provide higher
    #: overall progress.)
    #: This can be set in a subclass if not passed to the constructor.
    sleep = None

    #: How many total attempts will be made, including the initial. This
    #: can be set in a subclass, if not passed to the constructor.
    attempts = 3

    #: Commits longer than this (seconds) will trigger a warning log message.
    #: This can be set in a subclass if not passed to the constructor.
    long_commit_duration = DEFAULT_LONG_RUNNING_COMMIT_IN_SECS  # seconds

    transaction_sample_rate = 0.2

    #: The default return value from :meth:`should_abort_due_to_no_side_effects`.
    #: If you are not subclassing, or you do not need access to the arguments
    #: of the called function to make this determination, you may set this
    #: as an instance variable.
    side_effect_free = False

    #: If the number of resources joined to a transaction exceeds this,
    #: only a summary will be logged.
    #:
    #: .. versionadded:: 4.1.0
    side_effect_free_resource_report_limit = 5

    #: The log level at which to report violations of side-effect-free transactions
    #: (those that are reported as :meth:`should_abort_due_to_no_side_effects`,
    #: but which nonetheless have resource managers joined). This usually signifies
    #: a programming error, and results in throwing away work.
    #: If this is set to :obj:`logging.ERROR` or higher, an exception will be
    #: raised when this happens; this is useful for testing.
    #: The default value is :obj:`logging.DEBUG`.
    #:
    #: .. versionadded:: 4.1.0
    side_effect_free_log_level = DEBUG

    def __init__(
            self, handler,
            retries:int=None,
            sleep:float=None,
            long_commit_duration:float=None,
            transaction_manager=None,
    ): # pylint:disable=too-many-positional-arguments
        """
        :keyword float sleep: Sets the :attr:`sleep`.
        :keyword int retries: If given, the number of times a transaction will be
            retried. Note that this is one less than the total number of
            :attr:`attempts`
        :keyword transaction_manager: If not None, this is what will be returned
           from :meth:`get_transaction_manager_for_call` instead of the default
           transaction manager.

        .. versionchanged:: 5.0.0
           Add the *transaction_manager* argument.
        """
        self.handler = handler
        if retries is not None:
            self.attempts = retries + 1
        if long_commit_duration is not None:
            self.long_commit_duration = long_commit_duration
        if sleep is not None:
            self.sleep = sleep
            self.random = random.SystemRandom()
        self._txm = transaction_manager

    def __repr__(self):
        return '<%s.%s at 0x%x attempts=%s long_commit_duration=%s sleep=%s handler=%r>' % (
            type(self).__module__,
            type(self).__name__,
            id(self),
            self.attempts,
            self.long_commit_duration,
            self.sleep,
            self.handler
        )

    def prep_for_retry(self, attempts_remaining, tx, *args, **kwargs):
        """
        Called just after a transaction begins if there will be
        more than one attempt possible. Do any preparation
        needed to cleanup or prepare reattempts, or raise
        :class:`AbortAndReturn` if that's not possible.

        :param int attempts_remaining: How many attempts remain. Will always be
          at least 1.
        :param tx: The transaction that's just begun.

        .. versionchanged:: 3.1
           Add the *tx* parameter.
        """

    def should_abort_due_to_no_side_effects(self, *args, **kwargs): # pylint:disable=unused-argument
        """
        Called after the handler has run. If the handler should
        have produced no side effects and the transaction can be aborted
        as an optimization, return True.

        This defaults to the value of :attr:`side_effect_free`.
        """
        return self.side_effect_free

    def should_veto_commit(self, result, *args, **kwargs): # pylint:disable=unused-argument
        """
        Called after the handler has run. If the result of the handler
        should abort the transaction, return True.
        """
        return False

    _UNUSED_DESCRIPTION = 'Unknown'

    def describe_transaction(self, *args, **kwargs): # pylint:disable=unused-argument
        """
        Return a note for the transaction.

        This should return a string or None. If it returns a string,
        that value will be used via ``transaction.note()``
        """
        return self._UNUSED_DESCRIPTION

    def run_handler(self, *args, **kwargs):
        """
        Actually execute the callable handler.

        This is called from our ``__call__`` method. Subclasses
        may override to customize how the handler is called.
        """
        return self.handler(*args, **kwargs)

    #: Subclasses can customize this to a sequence of (Type, predicate)
    #: objects. If the transaction machinery doesn't know that an exception
    #: is retried, then we check in this list, checking for to see if it is an instance
    #: and applying the relevant test (which defaults to True)
    _retryable_errors = ()

    def _retryable(self, tx, exc_info): # pylint:disable=inconsistent-return-statements
        """
        Should the given exception info be considered one
        we should retry?

        By default, we consult the transaction manager, along with the
        list of (type, predicate) values we have in this object's
        ``_retryable_errors`` tuple.
        """
        v = exc_info[1]
        retryable = False
        try:
            retryable = tx.isRetryableError(v)
            if retryable:
                return retryable
        except Exception: # pylint:disable=broad-except
            pass
        else:
            # retryable was false
            for error_type, test in self._retryable_errors:
                if isinstance(v, error_type):
                    retryable = True if test is None else test(v)
                    break
            return retryable
        finally:
            del exc_info
            del v

    def _abort_on_exception(self, exc_info, retryable, number, tx):
        e = exc_info[0] # (t, v, tb)
        # There can be lots of valuable information in the str of
        # an exception instance. e.g., ZODB.POSException.ConflictError,
        # when raised by ZODB.ConflictResolation, contains the class name
        # and the OID involved, which is useful for tracking down
        # hotspots.
        try:
            tx.nti_abort()
            logger.debug("Transaction aborted; retrying %s/%s; %r: %s",
                         retryable, number, e, exc_info[1])
        except (AttributeError, ValueError):
            try:
                logger.exception("Failed to abort transaction following exception "
                                 "(retrying %s/%s; %r: %s). New exception:",
                                 retryable, number, e, exc_info[1])
            except:  # pylint:disable=I0011,W0702
                pass
            # We've seen RelStorage do this:
            # relstorage.cache:427 in after_poll:
            #    AttributeError: 'int' object has no attribute 'split'
            # which looks like
            # an issue with how it stores checkpoints in memcache.
            # We have no idea what state it's in after that, so we should
            # abort.

            # We've seen repoze.sendmail do this:
            # repoze.sendmail.delivery:119 in abort: ValueError "TPC in progress"
            # This appears to happen due to some other component raising an exception
            # after the call to commit has begun, and some exception slips through
            # such that, instead of calling `tpc_abort`, the stack unwinds.
            # The sendmail object appears to have been `tpc_begin`, but not
            # `tpc_vote`. (This may happen if the original exception was a ReadConflictError?)
            # https://github.com/repoze/repoze.sendmail/issues/31 (introduced in 4.2)
            # Again, no idea what state things are in, so abort with prejudice.
            try:
                if format_exception is not None:
                    fmt = format_exception(*exc_info)
                    logger.warning("Failed to abort transaction following exception. "
                                   "Original exception:\n%s",
                                   '\n'.join(fmt))
            except: # pylint:disable=bare-except
                exc_info = sys.exc_info()

            six.reraise(AbortFailedError, AbortFailedError(repr(exc_info)), exc_info[2])
        finally:
            del exc_info
            del e

    @staticmethod
    def setUp():
        """
        Called by `__call__` before making any attempts or beginning
        any transaction.

        When this method is called, it is guaranteed that ``transaction.manager.explicit``
        is true.

        Subclasses may override this method. If they are not a direct subclass
        of this class, they should be sure to call the `super` implementation; it
        is not necessary to call this implementation.

        .. versionadded:: 3.0
        """


    @staticmethod
    def tearDown():
        """
        Called by :meth:`__call__` just before returning, in all cases,
        once ``setUp`` has been called.

        When this method is called, it is guaranteed that ``transaction.manager.explicit``
        is at its original value.

        Subclasses may override this method. If they are not a direct subclass
        of this class, they should be sure to call the `super` implementation; it
        is not necessary to call this implementation.

        If this method raises an exception, the original return value of the handler,
        or its exception, will be lost.

        .. versionadded:: 3.0
        """

    _statsd_client = _statsd_client

    class _StatCollector(object):
        # Encapsulates the calls we make to send stats.
        __slots__ = ('client', 'rate', 'buf')
        def __init__(self, client, rate):
            self.client = client
            self.rate = rate
            # We will always send at least two stats, so we can save some network traffic
            # using a buffer. We shouldn't ever send enough to overflow a UDP packet
            # (based on napkin math)
            self.buf = []

        def __call__(self, name, count=1):
            self.client.incr(name, count=count, buf=self.buf, rate=self.rate)

        def flush(self):
            self.client.sendbuf(self.buf)
            del self.buf

    class _NullStatCollector(object):
        @staticmethod
        def __call__(name, count=1):
            """Does nothing"""

        @staticmethod
        def flush():
            """Does nothing"""

    _null_stat_collector = _NullStatCollector()

    # pylint:disable-next=unused-argument
    def get_transaction_manager_for_call(self, *args, **kwargs):
        """
        Return the transaction manager to use for a particular call.

        By default, this is the global/default thread-local transaction manager.

        The arguments are the arguments passed to ``__call__``.

        Subclasses may override. The expected use case is for transactions that are
        isolated and scoped to a particular well-defined location. Most
        likely, you'll keep a reference to the transaction manager outside
        the loop object (or, if the arguments don't matter, you can just call this method
        without supplying them)
        for later use. For example, when getting a connection from ZODB::

           txm = TransactionManager()
           loop = TransactionLoop(..., transaction_manager=txm)
           db = ... # get ZODB DB
           with db.open(loop.get_transaction_manager_for_call()) as conn:
               ...

        .. versionadded:: 5.0.0
        """
        # We use the thread-local global/default transaction manager.
        # Accessing it directly is a bit faster than going through the wrapping
        # layer. Applications should not be changing it.
        return self._txm if self._txm is not None else transaction.manager.manager

    def __call__(self, *args, **kwargs):
        note = self.describe_transaction(*args, **kwargs)

        # We use the thread-local global/default transaction manager.
        # Accessing it directly is a bit faster than going through the wrapping
        # layer. Applications should not be changing it.
        txm = self.get_transaction_manager_for_call(*args, **kwargs)
        # We always operate in explicit mode. This makes finding
        # errors such as committing or aborting multiple times much easier,
        # and prevents us from having to worry about a mis-match between
        # our local Transaction object and the global state.
        was_explicit = txm.explicit
        txm.explicit = True
        client = self._statsd_client()
        stats = (
            self._StatCollector(client, self.transaction_sample_rate)
            if client
            else self._null_stat_collector
        )

        try:
            self.setUp()
            return self.__loop(txm, note, stats, args, kwargs)
        except Exception:
            stats('transaction.failed')
            stats('transaction.retry', self.attempts - 1)
            raise
        finally:
            txm.explicit = was_explicit
            self.tearDown()
            stats.flush()

    def on_begin_failed(self, exc_info, txm, args, kwargs):
        """
        Called when ``begin()`` raised an error other than
        ``AlreadyInTransaction``, probably due to a bad synchronizer.
        Subclasses may override this method.

        After this is called, the tranasction manager *txm* will be aborted.

        This method must not raise an exception.

        .. versionadded:: 4.0.1
        """

    #: The event to send before running the handler the first time.
    #: Subclasses may override.
    #:
    #: ..versionadded:: 4.2.0
    EVT_WILL_FIRST_ATTEMPT = WillFirstAttempt
    #: The event to send before running the handler the second time,
    #: up until the last time.
    #: Subclasses may override.
    #:
    #: ..versionadded:: 4.2.0
    EVT_WILL_RETRY_ATTEMPT = WillRetryAttempt
    #: The event to send before running the handler the last time.
    #: Subclasses may override.
    #:
    #: ..versionadded:: 4.2.0
    EVT_WILL_LAST_ATTEMPT = WillLastAttempt

    # Where actually are we not using a return statement? ?
    # pylint:disable-next=too-many-positional-arguments,inconsistent-return-statements
    def __loop(self, txm, note, stats, args, kwargs):
        # pylint:disable=too-many-branches,too-many-statements,too-many-locals,too-complex

        attempts_remaining = self.attempts
        need_retry = self.attempts > 1
        begin = txm.begin
        unused_descr = self._UNUSED_DESCRIPTION
        sleep_time = 0
        invocation = LoopInvocation(self, self.handler, args, kwargs)
        while attempts_remaining:
            attempts_remaining -= 1
            # Starting at 0 for convenience
            attempt_number = self.attempts - attempts_remaining - 1

            # Bad synchronizers could cause this to raise.
            # That's not good.
            try:
                tx = begin()
            except AlreadyInTransaction: # pragma: no cover
                raise
            except:
                self.on_begin_failed(sys.exc_info(), txm, args, kwargs)

                try:
                    txm.abort()
                except Exception: # pylint:disable=broad-except
                    logger.exception(
                        "Failure when aborting transaction, after failure to begin transaction. "
                        "This exception will be suppressed in favor of the beginning exception "
                        "on Python 3; on Python 2, the beginning exception will be suppressed "
                        "in favor of this one."
                    )

                raise

            if note and note is not unused_descr:
                tx.note(note)
            notify(AfterTransactionBegan(invocation, tx))

            try:
                if need_retry:
                    self.prep_for_retry(attempts_remaining, tx, *args, **kwargs)

                if not attempt_number:
                    evt_kind = self.EVT_WILL_FIRST_ATTEMPT
                else:
                    evt_kind = (self.EVT_WILL_RETRY_ATTEMPT
                                if attempts_remaining
                                else self.EVT_WILL_LAST_ATTEMPT)

                notify(evt_kind(invocation, tx, attempt_number))

                result = self.run_handler(*args, **kwargs)

                # If the application called commit() or abort(), this will return None
                # A previous call to begin() to change the transaction would have raised
                # AlreadyInTransaction if the application hadn't committed or aborted
                # the transaction, which it should not of course be doing. If we don't check
                # this, then committing the transaction will raise a ValueError
                # from the TransactionManager: ValueError("Foreign transaction") which
                # happens *after* the transaction object has committed; not good.
                # Raise the error now so we can abort the proper object.
                current_tx = self.__has_current_transaction(txm)
                if current_tx is None:
                    tx = None
                    raise TransactionLifecycleError(
                        "The handler aborted or committed one or many transactions "
                        "and did not begin another one. Handlers must not perform "
                        "transaction lifecycle operations."
                    )
                if tx is not current_tx:
                    # Note that we don't handle the NoTransaction case, because
                    # we have no way of knowing whether the transaction was committed
                    # or aborted. Safer just to require the application not to manage
                    # its own transaction.
                    raise ForeignTransactionError(
                        "Transaction currently in progress is not the one the "
                        "loop began. It must have been committed and a new one started. "
                        "Handlers must not perform transaction lifecycle operations."
                    )

                if self.should_abort_due_to_no_side_effects(*args, **kwargs):
                    # These transactions can safely be aborted and
                    # ignored, reducing contention on commit
                    # locks, if any resources had actually been
                    # joined (ZODB Connection only joins when it
                    # detects a write, but those can later be undone by setting
                    # _p_changed = False; this doesn't un-join the transaction).

                    # NOTE: We raise these as an exception instead
                    # of aborting in the loop so that we don't
                    # retry if something goes wrong aborting
                    stats('transaction.side_effect_free')
                    # Detect if we are potentially throwing away work. The transaction
                    # was nominally side-effect free, but resource managers
                    # joined it, so they probably want to do work.
                    # This uses transaction internals, and we don't do a three-arg
                    # getattr(); we want to break if they change.
                    # The list of joined resource managers may be very long
                    # in some cases and it's not always useful to print
                    # all of them.
                    # pylint:disable=protected-access
                    resources = tx._resources or ()

                    if resources:
                        resources = (
                            '(count=%d)' % len(resources)
                            if len(resources) > self.side_effect_free_resource_report_limit
                            else resources
                        )

                        stats('transaction.side_effect_free_violation')
                        logger.log(
                            self.side_effect_free_log_level,
                            "Transaction %r nominally side-effect free has resource managers %s.",
                            tx, resources
                        )
                        if self.side_effect_free_log_level >= ERROR:
                            # Need to grab a copy of `resources` because this will clear it.
                            if not isinstance(resources, str):
                                resources = repr(resources)
                            self.__abort_transaction_quietly(tx)
                            raise TransactionLifecycleError(
                                "Transaction that was supposed to be side-effect free had "
                                "resource managers %s." % (resources,)
                            )
                    raise self.AbortAndReturn(result, "side-effect free")

                if tx.isDoomed() or self.should_veto_commit(result, *args, **kwargs):
                    stats('transaction.doomed' if tx.isDoomed() else 'transaction.vetoed')
                    raise self.AbortAndReturn(result, "doomed or vetoed")

                _do_commit(
                    tx,
                    self.long_commit_duration,
                    attempt_number,
                    sleep_time,
                )
                stats('transaction.successful')
                if attempt_number:
                    stats('transaction.retry', self.attempts - 1 - attempts_remaining)
                return result
            except ForeignTransactionError:
                # They left a transaction hanging around. If it's
                # still ACTIVE, we need to abort it, and clean up
                # after ourself if that fails, pending
                # https://github.com/zopefoundation/transaction/pull/84
                # This could raise lots of exceptions, including
                # ValueError(foreign transaction). We want to raise the FTE,
                # not an AbortFailedError.

                # Our current transaction, by definition, has already been
                # successfully committed or aborted. We only need to worry about
                # the new one, which is in an undetermined state.
                self.__abort_current_transaction_quietly(txm)
                raise

            except (
                    # Programming error: The application called
                    # commit() or abort() and then used transaction.get() (not us, we must
                    # never use an unguarded transaction.get()). The application
                    # should not call commit() or abort() and must be
                    # fixed. The good new is there's nothing to abort: by definition,
                    # our current transaction has been moved past.
                    NoTransaction,
                    # Programming error: The application called commit() or abort(),
                    # and we discovered that there was no active transaction.
                    # The good news is there's nothing to abort, as in the above.
                    TransactionLifecycleError
            ):
                raise
            except (
                    # Programming error: the application called begin() again.
                    # This should be fixed.
                    # The current transaction could still be our initial
                    # transaction, or it could be something else if they also had
                    # one of the other errors, so we have up to two transactions to
                    # to abort.
                    AlreadyInTransaction,
            ):
                current_tx = txm.get()
                self.__abort_current_transaction_quietly(txm)
                if current_tx is not tx:
                    self.__abort_transaction_quietly(tx)
                raise
            except self.AbortAndReturn as e:
                tx.nti_abort()
                return e.response
            except Exception: # pylint:disable=broad-except
                sleep_time += self.__handle_generic_exception(tx, attempts_remaining, invocation)
            except SystemExit:
                self.__handle_exit(tx)
                raise # pragma: no cover


    def __abort_current_transaction_quietly(self, txm):
        self.__abort_transaction_quietly(txm)
        # Ensure we're in a clean state
        # even if abort failed.
        # pylint:disable=protected-access
        txm._txn = None

    @staticmethod
    def __abort_transaction_quietly(tx):
        try:
            tx.abort()
        except Exception: # Ignore. pylint:disable=broad-except
            pass

    @staticmethod
    def __has_current_transaction(txm):
        # Handles an explicit transaction manager raising an exception
        # when it doesn't have a transaction.
        try:
            return txm.get()
        except NoTransaction:
            return None

    def __handle_generic_exception(self, tx, attempts_remaining, invocation,
                                   _reraise=six.reraise, _exc_info=sys.exc_info):
        # The code in the transaction package checks the retryable state
        # BEFORE aborting the current transaction. This matters because
        # aborting the transaction changes the transaction that the manager
        # has to a new one, and thus changes the set of registered resources
        # that participate in _retryable, depending on what synchronizers
        # are registered.
        exc_info = _exc_info()
        try:
            retryable = self._retryable(tx, exc_info)
            self._abort_on_exception(exc_info, retryable, attempts_remaining, tx)

            if attempts_remaining <= 0 or not retryable:
                _reraise(*exc_info)
        finally:
            exc_info = None

        if self.sleep:
            attempt_num = self.attempts - attempts_remaining # starting at 1
            backoff = self.random.randint(0, 2 ** attempt_num - 1)
            sleep_time = self.sleep * backoff
            event = WillSleepBetweenAttempts(invocation, sleep_time)
            notify(event)
            sleep_time = event.sleep_time
            self._sleep(sleep_time) # pylint:disable=too-many-function-args
        else:
            sleep_time = 0
        return sleep_time


    def __handle_exit(self, tx,
                      _reraise=six.reraise, _exc_info=sys.exc_info, _print_exc=print_exception):
        # If we are exiting, or otherwise probably going to
        # exit, do try to abort the transaction. The state of
        # the system is somewhat undefined at this point,
        # though, so don't try to time or log it, just print
        # to stderr on exception. Be sure to reraise the
        # original SystemExit.
        exc_info = _exc_info()
        try:
            try:
                tx.abort()
            except: # pylint:disable=bare-except
                _print_exc(*_exc_info())

            _reraise(*exc_info)
        finally:
            exc_info = None

    @Lazy
    def _sleep(self):
        # Delay accessing the sleep function until it's used in case
        # it gets monkey patched (e.g., gevent)
        return time.sleep



# -*- coding: utf-8 -*-
"""
Tests for pyramid_tween.py.

"""

import json
import unittest
import sys


from zope.component import getGlobalSiteManager
from transaction.interfaces import TransientError

from hamcrest import assert_that
from hamcrest import is_
from hamcrest import none
from hamcrest import has_entries
from hamcrest import contains_exactly
from hamcrest import is_not as does_not
from hamcrest import same_instance

from nti.testing.matchers import is_true
from nti.testing.matchers import is_false
from nti.testing.matchers import has_attr
from nti.testing.matchers import has_length
from nti.testing.matchers import validly_provides

from .. import pyramid_tween
from .._httpexceptions import HTTPBadRequest

# pylint:disable=protected-access,unnecessary-lambda-assignment

class MockRequest(object):

    method = 'GET'
    url = '/foo/bar'
    content_type = 'text/plain'
    body = None
    path_info = None
    make_body_seekable = lambda _: None

    def __init__(self, url=None):
        self.environ = {}
        self.url = url or self.url

class MockResponse(object):

    status = '200 OK'

    def __init__(self, headers=None, status=None):
        self.headers = {}
        self.headers.update(headers or {})
        self.status = status or self.status


class TestCommitVeto(unittest.TestCase):

    def _assert_aborts(self, req=None, rsp=None):
        assert_that(
            pyramid_tween.commit_veto(req, rsp),
            is_true()
        )

    def test_x_tm_abort(self):
        rsp = MockResponse(headers={'x-tm': 'abort'})
        self._assert_aborts(rsp=rsp)

    def test_x_tm_arbitrary(self):
        rsp = MockResponse(headers={'x-tm': 'anything'})
        self._assert_aborts(rsp=rsp)

    def test_status_4(self):
        rsp = MockResponse(status='4')
        self._assert_aborts(rsp=rsp)

    def test_status_5(self):
        rsp = MockResponse(status='5')
        self._assert_aborts(rsp=rsp)

    def test_x_tm_commit_overrides(self):
        rsp = MockResponse(headers={'x-tm': 'commit'}, status='500 Internal Error')
        assert_that(
            pyramid_tween.commit_veto(None, rsp),
            is_false()
        )


class TestSideEffectFree(unittest.TestCase):
    def _assert_free(self, req, free=True):
        assert_that(
            pyramid_tween.is_side_effect_free(req),
            is_true() if free else is_false()
        )

    def test_get(self):
        self._assert_free(MockRequest())

    def test_socketio_not_free(self):
        self._assert_free(
            MockRequest(url='socket.io'),
            free=False
        )
        self._assert_free(
            MockRequest(url='/socket.io/1/xhr-polling/0x2737c2a4c6b0cb4b?t=1574743462760'),
            free=False,
        )
        self._assert_free(
            MockRequest(url='prefix/socket.io/1/xhr-polling/0x2737c2a4c6b0cb4b?t=1574743462760'),
            free=False,
        )

    def test_static_socketio_free(self):
        self._assert_free(MockRequest(url='socket.io/static'))
        # Actual URL
        self._assert_free(
            MockRequest(url='/socket.io/static/socket.io.js')
        )
        # Prefix
        self._assert_free(
            MockRequest(url='prefix/socket.io/static/socket.io.js')
        )


class TestTransactionTween(unittest.TestCase):

    def _makeOne(self, handler=lambda request: MockResponse()):
        return pyramid_tween.TransactionTween(handler)

    def test_retry_attempts_in_environ(self):
        environs = []
        def handler(request):
            environs.append(request.environ.copy())
            raise TransientError

        loop = self._makeOne(handler)
        req = MockRequest()
        try:
            loop(req)
        except TransientError:
            pass
        assert_that(environs, has_length(3))
        assert_that(environs[0], has_entries('retry.attempt', 0,
                                             'retry.attempts', 3))
        assert_that(environs[1], has_entries('retry.attempt', 1,
                                             'retry.attempts', 3))
        assert_that(environs[2], has_entries('retry.attempt', 2,
                                             'retry.attempts', 3))

    def test_retry_attempts_reset_v_props(self):
        environs = []
        def handler(request):
            if not environs:
                # First time in.
                request._non_transient_private = 42
                request.non_transient_public = 24
                # It wasn't reset before beginning
                assert_that(request, has_attr('_v_from_before', 36))
                del request._v_from_before
            environs.append(request.environ.copy())
            assert_that(request, does_not(has_attr('_v_foo1')))
            assert_that(request, does_not(has_attr('_v_foo2')))
            assert_that(request, does_not(has_attr('_v_from_before')))
            assert_that(request, has_attr('_non_transient_private', 42))
            assert_that(request, has_attr('non_transient_public', 24))
            request._v_foo1 = 1
            request._v_foo2 = 2

            raise TransientError

        loop = self._makeOne(handler)
        req = MockRequest()
        req._v_from_before = 36 # pylint:disable=attribute-defined-outside-init
        try:
            loop(req)
        except TransientError:
            pass
        assert_that(environs, has_length(3))
        assert_that(environs[0], has_entries('retry.attempt', 0,
                                             'retry.attempts', 3))
        assert_that(environs[1], has_entries('retry.attempt', 1,
                                             'retry.attempts', 3))
        assert_that(environs[2], has_entries('retry.attempt', 2,
                                             'retry.attempts', 3))

    def test_retry_attempts_events(self):
        # pylint:disable=import-outside-toplevel
        # pylint:disable=too-many-locals
        from zope.event import subscribers

        from nti.transactions.interfaces import IAfterTransactionBegan
        from nti.transactions.interfaces import IWillFirstAttempt
        from nti.transactions.interfaces import IWillRetryAttempt
        from nti.transactions.interfaces import IWillLastAttempt

        from nti.transactions.interfaces import AfterTransactionBegan
        from nti.transactions.interfaces import WillFirstAttempt
        from nti.transactions.interfaces import WillRetryAttempt
        from nti.transactions.interfaces import WillLastAttempt

        from nti.transactions.interfaces import IWillFirstAttemptWithRequest
        from nti.transactions.interfaces import IWillRetryAttemptWithRequest
        from nti.transactions.interfaces import IWillLastAttemptWithRequest

        from nti.transactions.interfaces import WillFirstAttemptWithRequest
        from nti.transactions.interfaces import WillRetryAttemptWithRequest
        from nti.transactions.interfaces import WillLastAttemptWithRequest

        events = []
        subscribers.append(events.append)
        self.addCleanup(subscribers.remove, events.append)

        calls = []
        def handler(_):
            calls.append(1)
            raise TransientError

        loop = self._makeOne(handler)
        req = MockRequest()
        try:
            loop(req)
        except TransientError:
            pass
        assert_that(calls, has_length(3))

        assert_that(events, has_length(6))
        assert_that(events, contains_exactly(
            is_(AfterTransactionBegan),
            is_(WillFirstAttempt),

            is_(AfterTransactionBegan),
            is_(WillRetryAttempt),

            is_(AfterTransactionBegan),
            is_(WillLastAttempt),
        ))

        assert_that(events, contains_exactly(
            validly_provides(IAfterTransactionBegan),
            validly_provides(IWillFirstAttempt),

            validly_provides(IAfterTransactionBegan),
            validly_provides(IWillRetryAttempt),

            validly_provides(IAfterTransactionBegan),
            validly_provides(IWillLastAttempt),
        ))

        assert_that(events, contains_exactly(
            is_(AfterTransactionBegan),
            is_(WillFirstAttemptWithRequest),

            is_(AfterTransactionBegan),
            is_(WillRetryAttemptWithRequest),

            is_(AfterTransactionBegan),
            is_(WillLastAttemptWithRequest),
        ))

        assert_that(events, contains_exactly(
            validly_provides(IAfterTransactionBegan),
            validly_provides(IWillFirstAttemptWithRequest),

            validly_provides(IAfterTransactionBegan),
            validly_provides(IWillRetryAttemptWithRequest),

            validly_provides(IAfterTransactionBegan),
            validly_provides(IWillLastAttemptWithRequest),
        ))

        for i in (1, 3, 5):
            assert_that(events[i], has_attr(
                'request',
                is_(same_instance(req))))

    def test_is_last(self):
        is_lasts = []
        def handler(request):
            is_lasts.append(pyramid_tween.is_last_attempt(request))
            raise TransientError

        loop = self._makeOne(handler)
        req = MockRequest()
        try:
            loop(req)
        except TransientError:
            pass
        assert_that(is_lasts, has_length(3))
        assert_that(is_lasts, is_([False, False, True]))

    def test_is_retryable_true(self):
        is_retryable = []
        def handler(request):
            is_retryable.append(pyramid_tween.is_error_retryable(request, TransientError()))
            raise TransientError

        loop = self._makeOne(handler)
        req = MockRequest()
        try:
            loop(req)
        except TransientError:
            pass
        assert_that(is_retryable, has_length(3))
        assert_that(is_retryable, is_([True, True, False]))

    def test_is_retryable_false(self):
        is_retryable = []
        def handler(request):
            is_retryable.append(pyramid_tween.is_error_retryable(request, Exception()))

        loop = self._makeOne(handler)
        req = MockRequest()
        loop(req)
        assert_that(is_retryable, has_length(1))
        assert_that(is_retryable, is_([False]))


    def test_prep_for_retry_abort_on_IOError(self):
        loop = self._makeOne()
        class Request(MockRequest):
            def make_body_seekable(self):
                raise IOError

        with self.assertRaises(loop.AbortAndReturn) as exc:
            loop.prep_for_retry(None, None, Request())

        assert_that(exc.exception.response, is_(HTTPBadRequest))

    def test_prep_for_retry_replaces_content_type_of_list(self):
        req = MockRequest()
        req.content_type = 'application/x-www-form-urlencoded'
        req.body = json.dumps([]).encode('ascii')
        # GET is unchanged, body ignored.
        loop = self._makeOne()
        loop(req)
        assert_that(req.content_type, is_('application/x-www-form-urlencoded'))

        req.method = 'PUT'
        loop(req)
        assert_that(req.content_type, is_('application/json'))

    def test_prep_for_retry_replaces_content_type_of_dict(self):
        req = MockRequest()
        req.content_type = 'application/x-www-form-urlencoded'
        req.body = json.dumps({}).encode('ascii')

        loop = self._makeOne()
        loop(req)
        assert_that(req.content_type, is_('application/x-www-form-urlencoded'))

        req.method = 'POST'
        loop(req)
        assert_that(req.content_type, is_('application/json'))

    def test_should_abort_override(self):
        req = MockRequest()
        assert pyramid_tween.is_side_effect_free(req)

        loop = self._makeOne()
        assert_that(loop.should_abort_due_to_no_side_effects(req), is_true())

        req.environ['nti.request_had_transaction_side_effects'] = 42
        assert_that(loop.should_abort_due_to_no_side_effects(req), is_false())

    def test_run_handler_catches_httpexception(self):
        def handler(_req):
            raise HTTPBadRequest

        loop = self._makeOne(handler)
        req = MockRequest()
        res = loop.run_handler(req)

        assert_that(req, has_attr('_v_nti_raised_exception', True))
        assert_that(res, is_(HTTPBadRequest))

    def test_run_handler_throws(self):
        class MyExc(Exception):
            pass

        def handler(_req):
            raise MyExc

        loop = self._makeOne(handler)
        with self.assertRaises(MyExc):
            loop.run_handler(MockRequest())

    def test_call_raises_http_exception_when_raised(self):
        def handler(_req):
            raise HTTPBadRequest

        loop = self._makeOne(handler)
        with self.assertRaises(HTTPBadRequest):
            loop(MockRequest())

    def test_call_returns_http_exception_when_returned(self):
        def handler(_req):
            return HTTPBadRequest()

        loop = self._makeOne(handler)
        assert_that(loop(MockRequest()), is_(HTTPBadRequest))


class TestTransactionTweenFactory(unittest.TestCase):

    def setUp(self):
        try:
            from pyramid.config import Configurator
        except ModuleNotFoundError as ex:
            # pyramid 2.0.2 with webob 1.8.7 won't import on Python 3.13
            if sys.version_info[:2] == (3, 13):
                self.skipTest(str(ex))
        self.config = config = config = Configurator(registry=getGlobalSiteManager())
        config.setup_registry()

    def _makeOne(self, handler=None, **settings):
        self.config.registry.settings.update(settings)
        return pyramid_tween.transaction_tween_factory(handler, self.config.registry)

    def test_factory_simple(self):
        handler = 42

        loop = self._makeOne(handler)
        assert_that(loop, is_(pyramid_tween.TransactionTween))
        assert_that(loop, has_attr('handler', is_(handler)))
        assert_that(loop, has_attr('attempts', 3))
        assert_that(loop, has_attr('long_commit_duration', 6))
        assert_that(loop, has_attr('sleep', none()))

    def test_factory_attempts(self):
        loop = self._makeOne(**{'retry.attempts': 42})
        assert_that(loop.attempts, is_(42))

    def test_factory_long_commit(self):
        loop = self._makeOne(**{'retry.long_commit_duration': 42})
        assert_that(loop, has_attr('long_commit_duration', 42))

    def test_factory_sleep(self):
        loop = self._makeOne(**{'retry.sleep_ms': 10})
        assert_that(loop, has_attr('sleep', 0.01))



# -*- coding: utf-8 -*-
"""
Tests for the (deprecated) transactions.py.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import warnings

# pylint:disable=bad-option-value,import-outside-toplevel

class TestWarnings(unittest.TestCase):

    def test_add_abort_hook(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')

            import nti.transactions.transactions as T
            T.add_abort_hooks()

        self.assertEqual(len(w), 2, w)
        self.assertIn('is deprecated', str(w[0]))
        self.assertIn('test_transactions', str(w[0]))
        self.assertIn('add_abort_hooks', str(w[1]))
        self.assertIn('test_transactions', str(w[1]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
nti.transactions module.

Performs monkey patches when imported.
"""



from perfmetrics import Metric
from zope import interface

DEFAULT_LONG_RUNNING_COMMIT_IN_SECS = 6

###
# Monkey-patches
###

# By default, it wants to create a different logger
# for each and every thread or greenlet. We go through
# lots of greenlets, so that's lots of loggers
from transaction import _transaction # pylint:disable=import-private-name
# pylint:disable-next=protected-access
_transaction._LOGGER = __import__('logging').getLogger('txn.GLOBAL')

# Introduce a 'nti_abort' function that wraps the raw abort as a metric.
raw_abort = _transaction.Transaction.abort
raw_abort = getattr(raw_abort, 'im_func', raw_abort) # Py2
_transaction.Transaction.nti_abort = Metric('transaction.abort', rate=0.1)(raw_abort)
del raw_abort

# Ditto for commit
raw_commit = _transaction.Transaction.commit
raw_commit = getattr(raw_commit, 'im_func', raw_commit) # Py2
_transaction.Transaction.nti_commit = Metric('transaction.commit', rate=0.1)(raw_commit)
del raw_commit

from .interfaces import IExtendedTransaction
interface.alsoProvides(_transaction.Transaction, IExtendedTransaction)



# -*- coding: utf-8 -*-
"""
Constants to use for logging levels. See :mod:`ZODB.loglevels`.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


try:
    from ZODB.loglevels import TRACE
except ImportError: # pragma: no cover
    TRACE = 5

__all__ = [
    'TRACE',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Legacy module containing re-exports of other objects.
"""

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

logger = __import__('logging').getLogger(__name__)

# BWC imports
# For 4.0 makke this produce a deprecation warning. For 5.0
# remove this.
import warnings

from nti.transactions.loop import TransactionLoop
from nti.transactions.manager import ObjectDataManager
from nti.transactions.manager import OrderedNearEndObjectDataManager
from nti.transactions.manager import do
from nti.transactions.manager import do_near_end
from nti.transactions.queue import put_nowait

warnings.warn(
    "nti.transactions.transactions is deprecated.",
    FutureWarning,
    stacklevel=2
)

def add_abort_hooks():
    warnings.warn("add_abort_hooks is no longer needed.", FutureWarning, stacklevel=2)

__all__ = [
    'ObjectDataManager',
    'OrderedNearEndObjectDataManager',
    'TransactionLoop',
    'do',
    'do_near_end',
    'put_nowait',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

# disable: accessing protected members, too many methods
# pylint: disable=W0212,R0904

from hamcrest import is_
from hamcrest import has_length
from hamcrest import assert_that

import transaction
import six

try:
    from gevent import queue as gevent_queue
except ImportError:
    if six.PY2:
        from Queue import Full
        from Queue import Queue
    else:
        from queue import Full
        from queue import Queue
else: # pragma: no cover
    Full = gevent_queue.Full
    Queue = gevent_queue.Queue

from ..queue import put_nowait

from nti.testing.base import AbstractTestBase

class PutQueueTest(AbstractTestBase):

    def test_put_succeeds(self):
        queue = Queue() # unbounded
        transaction.begin()

        put_nowait(queue, self)
        # still empty
        assert_that(queue.qsize(), is_(0))

        transaction.commit()

        assert_that(queue.get(block=False), is_(self))

    def test_put_transaction_abort(self):
        queue = Queue()
        transaction.begin()
        put_nowait(queue, 'aborted')
        transaction.abort()

        transaction.begin()
        put_nowait(queue, 'committed')
        transaction.commit()

        assert_that(queue.qsize(), is_(1))
        assert_that(queue.get(block=False), is_('committed'))

    def test_put_transaction_savepoint(self):
        queue = Queue()
        transaction.begin()
        put_nowait(queue, 'presavepoint')
        # we can get a non-optimistic savepoint
        savepoint = transaction.savepoint(optimistic=False)
        assert_that(savepoint._savepoints, has_length(1))
        repr(savepoint._savepoints) # cover
        put_nowait(queue, 'aftersavepoint')

        # If we rollback the savepoint now, what we just
        # did will be lost, but the original work
        # will still happen
        savepoint.rollback()
        transaction.commit()

        assert_that(queue.qsize(), is_(1))
        assert_that(queue.get(block=False), is_('presavepoint'))

    def test_put_multiple_correct_order(self):
        # Early builds had a bug where the sort order of the datamanagers
        # was non-deterministic since it was based on object id, and that's not
        # guaranteed to be atomically increasing. It takes a high iteration count to
        # demonstrate this, though
        queue = Queue()

        for _ in range(10000):
            transaction.begin()

            put_nowait(queue, 'a')
            put_nowait(queue, 'b')

            transaction.commit()

            assert_that(queue.get(block=False), is_('a'))
            assert_that(queue.get(block=False), is_('b'))

    def test_put_failure(self):
        queue = Queue(1) # unbounded
        queue.put(object())
        assert_that(queue.qsize(), is_(1))

        transaction.begin()

        put_nowait(queue, self)
        # still size 1
        assert_that(queue.qsize(), is_(1))
        with self.assertRaises(Full) as cm:
            transaction.commit()


        assert_that(cm.exception, is_(Full))
        assert_that(queue.get(block=False), is_(object))



# -*- coding: utf-8 -*-
"""
Exception classes that represent HTTP-level status.

See :mod:`pyramid.httpexceptions`

"""

class HTTPException(Exception):
    """Placeholder if pyramid is not installed."""

class HTTPBadRequest(HTTPException):
    """Placeholder if pyramid is not installed."""

try:
    from pyramid import httpexceptions
except ImportError: # pragma: no cover
    pass
else:
    HTTPException = httpexceptions.HTTPException
    HTTPBadRequest = httpexceptions.HTTPBadRequest

__all__ = [
    'HTTPException',
    'HTTPBadRequest',
]



import codecs
from setuptools import setup
from setuptools import find_namespace_packages

version = '5.1.0'

entry_points = {
    'console_scripts': [
    ],
}

TESTS_REQUIRE = [
    'coverage',
    'nti.testing',
    'pylint',
    'pyramid',
    'zope.component',
    'zope.testrunner',
    'ZODB',
]

def _read(fname):
    with codecs.open(fname, encoding='UTF-8') as f:
        return f.read()

setup(
    name='nti.transactions',
    version=version,
    author='Jason Madden',
    author_email='jason@nextthought.com',
    description="NTI Transactions Utility",
    long_description=_read('README.rst') + _read('CHANGES.rst'),
    license='Apache',
    keywords='ZODB transaction',
    url='https://github.com/OpenNTI/nti.transactions',
    zip_safe=True,
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Framework :: ZODB',
    ],
    python_requires=">=3.10",
    packages=find_namespace_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'six',
        'perfmetrics',
        'transaction >= 3.0.0',
        'zope.cachedescriptors',
        'zope.exceptions',
        'zope.interface',
        'zope.event',
    ],
    extras_require={
        'test': TESTS_REQUIRE,
        'gevent': ['gevent',],
        'docs': [
            'Sphinx',
            'repoze.sphinx.autointerface',
            'pyhamcrest',
            'furo; python_version >= "3.6"',
            'sphinx_rtd_theme; python_version < "3.6"',
        ],
        'pyramid': [
            'pyramid >= 1.2',
        ],
    },
    entry_points=entry_points
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Support for data managers.
"""

from zope import interface

import transaction

from transaction.interfaces import IDataManagerSavepoint
from transaction.interfaces import ISavepointDataManager

__all__ = [
    'ObjectDataManager',
    'OrderedNearEndObjectDataManager',
    'do',
    'do_near_end',
]

@interface.implementer(ISavepointDataManager, IDataManagerSavepoint)
class ObjectDataManager(object):
    """
    A generic (and therefore relatively expensive)
    :class:`transaction.interfaces.IDataManager` that invokes a
    callable (usually a method of an object) when the transaction
    finishes successfully. The method should not raise exceptions when
    invoked, as they will be caught and ignored (to preserve
    consistency with the rest of the data managers). If there's a
    chance the method could fail, then whatever actions it does take
    should not have side-effects.

    These data managers have no guaranteed relationship to other data
    managers in terms of the order of which they commit, except as
    documented with :meth:`sortKey`.

    Because these data managers execute exactly one operation on a
    complete transaction commit, implementing a savepoint is trivial:
    do nothing when it is rolled back. A savepoint is created to checkpoint
    a transaction and rolled back to reverse actions taken *after* the
    checkpoint. Only data managers that were active (joined) at the
    time the transaction savepoint is created are asked to create
    their own savepoint, and then potentially to roll it back. We do
    no work until the transaction is committed, so we have nothing
    to rollback. Moroever, if a transaction savepoint is activated
    before a manager joins, then that manager is not asked for its own
    savepoint: it is simply aborted and unjoined from the transaction if
    the previous savepoint is rolledback.

    """

    _EMPTY_KWARGS = {}

    # pylint:disable-next=too-many-positional-arguments
    def __init__(self, target=None, method_name=None, call=None,
                 vote=None, args=(), kwargs=None):
        """
        Construct a data manager. You must pass either the `target` and `method_name` arguments
        or the `call` argument. (You may always pass the `target` argument, which will
        be made available on this object for the use of :meth:`tpc_vote`. )

        :param target: An object. Optional if `call` is given. If provided, will be used
            to compute the :meth:`sortKey`.
        :param string method_name: The name of the attribute to get from `target`.
            Optional if `callable` is given.
        :param callable call: A callable object. Ignored if `target` *and* `method_name` are given.
        :param callable vote: If given, then a callable that will be called during the voting phase.
            It should raise an exception if the transaction will fail.
        :param args: A sequence of arguments to pass to the callable object. Optional.
        :param kwargs: A dictionary of arguments to pass to the callable object. Optional.
        """
        self.target = target
        if method_name:
            self.callable = getattr(target, method_name)
        else:
            self.callable = call

        assert self.callable is not None

        self.args = args
        self.kwargs = kwargs or self._EMPTY_KWARGS

        self.vote = vote

        # Use the default thread transaction manager.
        self.transaction_manager = transaction.manager

    def commit(self, tx):
        pass

    def abort(self, tx):
        pass

    def sortKey(self):
        """
        Return the string value that, when sorted, determines the
        order in which data managers will get to vote and commit at
        the end of a transaction. (See
        :meth:`transaction.interfaces.IDataManager.sortKey`).

        The default implementation of this method uses the ID of
        either the ``target`` object we were initialized with or the ID of
        the actual callable we will call. This has the property of
        ensuring that *all* calls to methods of a particular object
        instance (when ``target`` is given), or calls to a particular callable
        (when ``target`` is not given) will execute in the order in which they were
        added to the transaction.

        .. note:: This relies on an implementation detail of the
            transaction package, which sorts using :meth:`list.sort`,
            which is guaranteed to be stable: thus objects using the
            same key remain in the same relative order. (See
            :meth:`transaction._transaction.Transaction._commitResources`.)

        To execute only calls to a particular method of a particular instance
        in the order they are added to the transaction, but allow other
        methods to execute before or after them, do not provide the ``target``.

        It is not advisable to use the ID of this object (``self``) in
        the implementation of this method, because the ID values are
        not guaranteed to be monotonically increasing and thus
        instances of a particular class that did this would execute in
        "random" order.
        """
        # It's not clearly documented, but this is supposed to be a string
        return str(id(self.target) if self.target is not None else id(self.callable))

    # No subtransaction support.
    def abort_sub(self, tx):
        """Does nothing"""

    commit_sub = abort_sub

    def beforeCompletion(self, tx):
        """Does nothing"""

    afterCompletion = beforeCompletion

    def tpc_begin(self, tx, subtransaction=False): # pylint:disable=unused-argument
        assert not subtransaction

    def tpc_vote(self, tx): # pylint:disable=unused-argument
        if self.vote:
            self.vote()

    def tpc_finish(self, tx): # pylint:disable=unused-argument
        self.callable(*self.args, **self.kwargs)

    tpc_abort = abort

    def __repr__(self):
        return '<%s.%s at %s for %r>' % (self.__class__.__module__, self.__class__.__name__,
                                         id(self),
                                         self.callable)

    # ISavepointDataManager
    def savepoint(self):
        return self

    # IDatamanagerSavepoint
    def rollback(self):
        # See class comments: we have nothing to rollback
        # from because we take no action until commit
        # anyway.
        pass

class OrderedNearEndObjectDataManager(ObjectDataManager):
    """
    A special extension of :class:`ObjectDataManager` that attempts to execute
    after all other data managers have executed. This is useful when an
    operation relies on the execution of other data managers.

    .. versionadded:: 1.1
    """

    def sortKey(self):
        """
        Sort prepended with z's in an attempt to execute after other data
        managers.
        """
        parent_key = super().sortKey()
        sort_str = str(self.target) if self.target is not None else str(self.callable)
        return 'zzz%s:%s' % (sort_str, parent_key)

def do(*args, **kwargs):
    """
    Establishes a IDataManager in the current transaction.
    See :class:`ObjectDataManager` for the possible arguments.
    """
    klass = kwargs.pop('datamanager_class', ObjectDataManager)
    result = klass(*args, **kwargs)
    transaction.get().join(result)
    return result

def do_near_end(*args, **kwargs):
    """
    Establishes a IDataManager in the current transaction that will attempt to
    execute *after* all other DataManagers have had their say.
    See :class:`ObjectDataManager` for the possible arguments.

    .. versionadded:: 1.1
    """
    kwargs['datamanager_class'] = OrderedNearEndObjectDataManager
    return do(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pylint:disable=too-many-public-methods
# pylint:disable=import-outside-toplevel

import logging
import sys
import unittest

import zope.event

from hamcrest import assert_that
from hamcrest import is_
from hamcrest import is_not as does_not
from hamcrest import calling
from hamcrest import raises
from hamcrest import has_property
from hamcrest import none
from hamcrest import has_items
from hamcrest import greater_than_or_equal_to
from hamcrest import contains_exactly as contains
from hamcrest import contains_string



from nti.testing.matchers import is_true
from nti.testing.matchers import is_false
from nti.testing.matchers import has_length
from nti.testing.matchers import validly_provides

from ..interfaces import CommitFailedError
from ..interfaces import AbortFailedError
from ..interfaces import ForeignTransactionError
from ..interfaces import TransactionLifecycleError
from ..interfaces import AfterTransactionBegan
from ..interfaces import WillFirstAttempt
from ..interfaces import WillRetryAttempt
from ..interfaces import WillSleepBetweenAttempts
from ..interfaces import IAfterTransactionBegan
from ..interfaces import IWillRetryAttempt
from ..interfaces import IWillSleepBetweenAttempts

from ..loop import _do_commit
from ..loop import TransactionLoop

import transaction
from transaction.interfaces import TransientError
from transaction.interfaces import NoTransaction
from transaction.interfaces import AlreadyInTransaction

from perfmetrics.testing import FakeStatsDClient
from perfmetrics.testing.matchers import is_counter
from perfmetrics import statsd_client_stack

from ZODB import DB
from ZODB.DemoStorage import DemoStorage
from ZODB.POSException import StorageError

from unittest import mock


# pylint:disable=protected-access,broad-exception-raised

class Test_Do_Commit(unittest.TestCase):
    class Transaction(object):
        description = ''
        def __init__(self, t=None):
            self.t = t

        def nti_commit(self):
            if self.t:
                raise self.t # (Python2, old pylint)  pylint:disable=raising-bad-type

    def RaisingCommit(self, t=Exception):
        return self.Transaction(t)

    def test_commit_raises_type_error_raises_commit_failed(self):
        assert_that(calling(_do_commit)
                    .with_args(
                        self.RaisingCommit(TypeError),
                        '', 0, 0, 0
                    ),
                    raises(CommitFailedError))

    def test_commit_raises_type_error_raises_commit_failed_good_message(self):
        assert_that(calling(_do_commit)
                    .with_args(
                        self.RaisingCommit(TypeError("A custom message")),
                        '', 0, 0, 0,
                    ),
                    raises(CommitFailedError, "A custom message"))


    @mock.patch('nti.transactions.loop.logger.exception')
    def test_commit_raises_assertion_error(self, fake_logger):
        assert_that(calling(_do_commit)
                    .with_args(
                        self.RaisingCommit(AssertionError), '', 0, 0, 0
                    ),
                    raises(AssertionError))
        fake_logger.assert_called_once()


    @mock.patch('nti.transactions.loop.logger.exception')
    def test_commit_raises_value_error(self, fake_logger):
        fake_logger.expects_call()

        assert_that(calling(_do_commit)
                    .with_args(
                        self.RaisingCommit(ValueError),
                        '', 0, 0, 0,
                    ),
                    raises(ValueError))

    @mock.patch('nti.transactions.loop.logger.exception')
    def test_commit_raises_custom_error(self, fake_logger):
        fake_logger.expects_call()

        class MyException(Exception):
            pass

        try:
            raise MyException()
        except MyException:
            assert_that(calling(_do_commit)
                        .with_args(
                            self.RaisingCommit(ValueError),
                            '', 0, 0, 0
                        ),
                        raises(MyException))

    @mock.patch('nti.transactions.loop.logger.log')
    def test_commit_clean_but_long(self, fake_logger):
        fake_logger.expects_call()
        _do_commit(self.RaisingCommit(None), -1, 0, 0)



    @mock.patch('nti.transactions.loop.logger.log')
    @mock.patch('nti.transactions.loop.logger.isEnabledFor')
    def test_commit_duration_logging_short(self, fake_is_enabled, fake_log):
        def enabled(lvl):
            return lvl == logging.DEBUG
        fake_is_enabled.side_effect = enabled

        _do_commit(self.Transaction(), 6, 0, 0)
        fake_log.assert_called_once()

    @mock.patch('nti.transactions.loop.logger.log')
    @mock.patch('nti.transactions.loop.logger.isEnabledFor')
    def test_commit_duration_logging_long(self, fake_is_enabled, fake_log):
        def enabled(lvl):
            return lvl == logging.WARNING
        fake_is_enabled.side_effect = enabled
        perfs = [10, 0]
        def perf():
            return perfs.pop()
        fake_perf_counter = mock.MagicMock()
        fake_perf_counter.side_effect = perf
        _do_commit(self.Transaction(), 6, 0, 0, _perf_counter=fake_perf_counter)
        fake_log.assert_called_once()

class TrueStatsDClient(FakeStatsDClient):
    # https://github.com/zodb/perfmetrics/issues/23
    def __bool__(self):
        return True
    __nonzero__ = __bool__


class TestLoop(unittest.TestCase):

    def setUp(self):
        try:
            transaction.abort()
        except NoTransaction: # pragma: no cover
            pass
        transaction.manager.clearSynchs()
        self.statsd_client = TrueStatsDClient()
        self.statsd_client.random = lambda: 0 # Ignore rate, capture all packets
        statsd_client_stack.push(self.statsd_client)
        self.events = []
        zope.event.subscribers.append(self.events.append)

    def tearDown(self):
        statsd_client_stack.pop()
        zope.event.subscribers.remove(self.events.append)
        transaction.manager.clearSynchs()

    @mock.patch('nti.transactions.loop._do_commit')
    def test_trivial(self, fake_commit):
        class Any(object):
            def __eq__(self, other):
                return True
            def __hash__(self):
                return 42

        loop = TransactionLoop(lambda a: a, retries=1, long_commit_duration=1, sleep=1)
        r = repr(loop)
        assert_that(r, contains_string('sleep=1'))
        assert_that(r, contains_string('long_commit_duration=1'))
        assert_that(r, contains_string('attempts=2'))
        fake_commit.expects_call().with_args(
            Any(), # transaction
            loop.long_commit_duration,
            0, # attempt number / retries
            0 # sleep_time
        )

        result = loop(1)
        assert_that(result, is_(1))
        # May or may not get a transaction.commit stat first, depending on random
        assert_that(self.statsd_client.packets, has_length(greater_than_or_equal_to(1)))
        assert_that(self.statsd_client.observations[-1],
                    is_counter(name='transaction.successful', value=1))

        assert_that(self.events, has_length(2))
        assert_that(self.events, contains(is_(AfterTransactionBegan), is_(WillFirstAttempt)))

    def test_explicit(self):
        assert_that(transaction.manager, has_property('explicit', is_false()))

        def handler():
            assert_that(transaction.manager, has_property('explicit', is_true()))
            return 42

        result = TransactionLoop(handler)()
        assert_that(result, is_(42))

    def test_synchronizer_raises_error_on_begin(self):
        class SynchError(Exception):
            pass

        class Synch(object):
            count = 0
            def newTransaction(self, _txn):
                self.count += 1
                if self.count == 1:
                    raise SynchError


            def afterCompletion(self, _txm):
                pass

            beforeCompletion = afterCompletion


        synch = Synch()
        transaction.manager.registerSynch(synch)

        class HandlerError(Exception):
            pass

        def handler():
            raise HandlerError

        # Doing it the first time fails
        loop = TransactionLoop(handler)
        with self.assertRaises(SynchError):
            loop()

        # Our synch doesn't raise the second time,
        # and we don't get AlreadyInTransaction.
        with self.assertRaises(HandlerError):
            loop()

    def test_zodb_synchronizer_raises_error_on_begin(self):
        # Closely mimic what we see in
        # https://github.com/NextThought/nti.transactions/issues/49,
        # where the storage's ``pollInvalidations`` method
        # raises errors.
        db = DB(DemoStorage())
        # The connection has to be open to register a synch
        conn = db.open()

        def bad_poll_invalidations():
            raise StorageError

        conn._storage.poll_invalidations = bad_poll_invalidations

        # For the fun of it, lets assume that afterCompletion is also broken
        class CompletionError(Exception):
            pass
        def bad_afterCompletion():
            raise CompletionError
        conn._storage.afterCompletion = bad_afterCompletion

        def handler():
            self.fail("Never get here") # pragma: no cover

        loop = TransactionLoop(handler)

        # Python 2 and Python 3 raise different things
        expected = StorageError if str is not bytes else CompletionError
        for _ in range(2):
            with self.assertRaises(expected):
                loop()

    def test_explicit_begin(self):
        def handler():
            transaction.begin()

        assert_that(calling(TransactionLoop(handler)), raises(AlreadyInTransaction))

    def test_explicit_begin_after_commit(self):
        # We change the current transaction out and then still manage to raise
        # AlreadyInTransaction
        def handler():
            transaction.abort()
            transaction.begin()
            transaction.begin()

        assert_that(calling(TransactionLoop(handler)), raises(AlreadyInTransaction))


    def test_explicit_end(self):
        def handler():
            transaction.abort()

        assert_that(calling(TransactionLoop(handler)), raises(TransactionLifecycleError))

    def test_explicit_foreign(self):
        def handler():
            transaction.abort()
            transaction.begin()

        assert_that(calling(TransactionLoop(handler)), raises(ForeignTransactionError))

    def test_explicit_foreign_abort_fails(self):
        def bad_abort():
            raise Exception("Bad abort")

        def handler():
            transaction.abort()
            tx = transaction.begin()
            tx.abort = tx.nti_abort = bad_abort

        assert_that(calling(TransactionLoop(handler)), raises(ForeignTransactionError))
        assert_that(transaction.manager.manager, has_property('_txn', is_(none())))

    def test_setup_teardown(self):

        class Loop(TransactionLoop):
            # false positive pylint:disable=arguments-differ
            setupcalled = teardowncalled = False
            def setUp(self):
                assert_that(transaction.manager, has_property('explicit', is_true()))
                self.setupcalled = True
            def tearDown(self):
                self.teardowncalled = True

        def handler():
            raise Exception

        loop = Loop(handler)
        assert_that(calling(loop), raises(Exception))

        assert_that(loop, has_property('setupcalled', is_true()))
        assert_that(loop, has_property('teardowncalled', is_true()))


    def _check_retriable(self,  loop_class=TransactionLoop, exc_type=TransientError,
                         *,
                         raise_count=1, loop_args=(), loop_kwargs=None):
        calls = []
        def handler():
            # exc_info should be clear on entry.
            assert_that(sys.exc_info(), is_((None, None, None)))
            if len(calls) < raise_count:
                calls.append(1)
                raise exc_type(calls)
            return "hi"

        loop = loop_class(handler, *loop_args, **(loop_kwargs or {}))
        result = loop()
        assert_that(result, is_("hi"))
        assert_that(calls, is_([1] * raise_count))
        observations = self.statsd_client.observations
        assert_that(observations,
                    has_items(
                        is_counter(name='transaction.successful', value=1),
                        is_counter(name='transaction.retry', value=raise_count)))
        return loop

    def test_retriable(self):
        self._check_retriable()

    def test_custom_retriable(self):
        class Loop(TransactionLoop):
            _retryable_errors = ((Exception, None),)

        self._check_retriable(Loop, AssertionError)

    def test_retriable_gives_up(self):
        def handler():
            raise TransientError()
        loop = TransactionLoop(handler, sleep=0.01, retries=1)
        assert_that(calling(loop), raises(TransientError))

    def test_non_retryable(self):
        class MyError(Exception):
            pass
        def handler():
            raise MyError()
        loop = TransactionLoop(handler, sleep=0.01, retries=100000000)
        assert_that(calling(loop), raises(MyError))
        observations = self.statsd_client.observations
        assert_that(observations,
                    has_items(
                        is_counter(name='transaction.failed', value=1)))

    def test_isRetryableError_exception(self):
        # If the transaction.isRetryableError() raises, for some reason,
        # we still process our list
        class MyError(object):
            pass
        class Loop(TransactionLoop):
            _retryable_errors = ((MyError, None),)

        loop = Loop(None)
        loop._retryable(None, (None, MyError(), None))

    def test_retryable_backoff(self):
        class NotRandom(object):
            def randint(self, _floor, ceiling):
                return ceiling

        class Loop(TransactionLoop):
            attempts = 10
            def __init__(self, *args, **kwargs):
                TransactionLoop.__init__(self, *args, **kwargs)
                self.times = []
                self.random = NotRandom()
                self._sleep = self.times.append

        # By default, it is not called.
        loop = self._check_retriable(Loop, raise_count=5)
        assert_that(loop, has_property('times', []))

        # Setting a delay calls it
        loop = self._check_retriable(Loop, raise_count=5, loop_kwargs={'sleep': 0.1})
        # The ceiling arguments are 2**attempt - 1, so
        # 1, 3, 7, 15, 31, and sleep times are
        # 0.1, 0.3, 0.7, 1.5, 3,1
        times = [(2 ** x - 1) * 0.1 for x in range(1, 6)]
        assert_that(loop, has_property('times',
                                       times))

        assert_that(self.events, has_length(29))

        assert_that(self.events[-1], is_(WillRetryAttempt))
        assert_that(self.events[-2], is_(AfterTransactionBegan))
        assert_that(self.events[-3], is_(WillSleepBetweenAttempts))
        assert_that(self.events[-1], validly_provides(IWillRetryAttempt))
        assert_that(self.events[-2], validly_provides(IAfterTransactionBegan))
        assert_that(self.events[-3], validly_provides(IWillSleepBetweenAttempts))

        assert_that(self.events[-3], has_property('sleep_time', 3.1))


    @mock.patch('transaction._manager.TransactionManager.get', autospec=True)
    @mock.patch('transaction._manager.TransactionManager.begin', autospec=True)
    def test_note(self, fake_begin, fake_get):
        fake_tx = mock.MagicMock()
        fake_tx.isDoomed.return_value = True
        fake_begin.return_value = fake_tx
        fake_get.return_value = fake_tx

        class Loop(TransactionLoop):
            def describe_transaction(self, *args, **kwargs):
                return "Hi"

        result = Loop(lambda: 42)()
        assert_that(result, is_(42))
        fake_tx.note.assert_called_with('Hi')
        fake_tx.nti_abort.assert_called_once()


    @mock.patch('transaction._manager.TransactionManager.begin', autospec=True)
    @mock.patch('transaction._manager.TransactionManager.get', autospec=True)
    def test_abort_no_side_effect(self, fake_get, fake_begin):
        fake_tx = mock.MagicMock()
        fake_tx._resources = () # pylint:disable=protected-access

        fake_begin.return_value = fake_tx
        fake_get.return_value = fake_tx


        class Loop(TransactionLoop):
            side_effect_free = True

        result = Loop(lambda: 42)()
        assert_that(result, is_(42))
        observations = self.statsd_client.observations
        assert_that(observations,
                    has_items(
                        is_counter(name='transaction.side_effect_free', value=1)))
        assert_that(observations,
                    does_not(
                        has_items(
                            is_counter(name='transaction.side_effect_free_violation')
                        )))

        fake_tx.nti_abort.assert_called_once()

    def test_abort_no_side_effect_violation(self):
        class Mock(mock.MagicMock):
            def __str__(self):
                return 'fake:fake_manager'
            __repr__ = __str__

        fake_manager = Mock()

        class Loop(TransactionLoop):
            side_effect_free = True

        def handler():
            transaction.get().join(fake_manager)

        loop = Loop(handler)
        loop()
        observations = self.statsd_client.observations
        assert_that(observations,
                    has_items(
                        is_counter(name='transaction.side_effect_free', value=1)))
        assert_that(observations,
                    has_items(
                        is_counter(name='transaction.side_effect_free_violation', value=1)
                    ))

        loop.side_effect_free_log_level = logging.ERROR
        with self.assertRaises(TransactionLifecycleError) as exc:
            loop()

        ex = exc.exception
        assert_that(str(ex),
                    is_("Transaction that was supposed to be side-effect free "
                        "had resource managers [fake:fake_manager]."))

        loop.side_effect_free_resource_report_limit = 0
        with self.assertRaises(TransactionLifecycleError) as exc:
            loop()

        ex = exc.exception
        assert_that(str(ex),
                    is_("Transaction that was supposed to be side-effect free "
                        "had resource managers (count=1)."))


    @mock.patch('transaction._transaction.Transaction.nti_abort')
    def test_abort_doomed(self, fake_abort):
        fake_abort.expects_call()

        def handler():
            assert_that(transaction.manager.explicit, is_true())
            transaction.get().doom()
            return 42

        result = TransactionLoop(handler)()
        assert_that(result, is_(42))
        observations = self.statsd_client.observations
        assert_that(observations,
                    has_items(
                        is_counter(name='transaction.doomed', value=1)))


    @mock.patch('transaction._manager.TransactionManager.get')
    @mock.patch('transaction._manager.TransactionManager.begin')
    def test_abort_veto(self, fake_begin, fake_get):
        fake_tx = mock.MagicMock()
        fake_tx.isDoomed.return_value = False

        fake_begin.return_value = fake_tx
        fake_get.return_value = fake_tx

        class Loop(TransactionLoop):
            def should_veto_commit(self, result, *args, **kwargs):
                assert_that(result, is_(42))
                return True

        result = Loop(lambda: 42)()
        assert_that(result, is_(42))
        observations = self.statsd_client.observations
        assert_that(observations,
                    has_items(
                        is_counter(name='transaction.vetoed', value=1)))

        fake_tx.nti_abort.assert_called_once()


    @mock.patch('transaction._manager.TransactionManager.begin', autospec=True)
    @mock.patch('sys.stderr', autospec=True)
    def test_abort_systemexit(self, fake_begin, _fake_stderr):
        fake_tx = mock.MagicMock()
        fake_tx.abort.side_effect = ValueError
        fake_tx.isDoomed.return_value = False

        fake_begin.return_value = fake_tx

        def handler():
            raise SystemExit()

        loop = TransactionLoop(handler)
        try:
            loop()
            self.fail("Should raise SystemExit") # pragma: no cover
        except SystemExit:
            pass


    @mock.patch('nti.transactions.loop.logger.exception', side_effect=ValueError)
    @mock.patch('nti.transactions.loop.logger.warning', side_effect=ValueError)
    @mock.patch('transaction._manager.TransactionManager.begin',)
    def test_abort_exception_raises(self, fake_begin,
                                    _fake_logger, _fake_format):
        # begin() returns an object without abort(), which we catch.
        # Likewise for the things we try to do to log it
        fake_begin.return_value = object()

        def handler():
            raise Exception()
        loop = TransactionLoop(handler)
        with self.assertRaises(AbortFailedError):


            loop()


    def test_abort_on_exception_logs_exception_str(self):
        from zope.testing.loggingsupport import InstalledHandler
        from ZODB.POSException import ConflictError
        import pickle
        logs = InstalledHandler("nti.transactions.loop")
        self.addCleanup(logs.uninstall)

        loop = TransactionLoop(lambda: None)

        pickle_data = pickle.dumps(type(self), 2)


        exc_info = (
            ConflictError,
            # Provide the data that ZODB.ConflictResolution does
            ConflictError(
                oid=b'\x01' * 8,
                serials=(b'\x02' * 8, b'\x01' * 8),
                data=pickle_data
            ),
            None # Traceback not used
        )

        class Tx(object):
            @staticmethod
            def nti_abort():
                """Nothing"""

        loop._abort_on_exception(exc_info, True, 4, Tx)

        self.assertEqual(len(logs.records), 1)
        msg = logs.records[0].getMessage()
        assert_that(msg, is_(
            "Transaction aborted; "
            "retrying True/4; "
            "<class 'ZODB.POSException.ConflictError'>: database conflict error "
            "(oid 0x0101010101010101, "
            "class nti.transactions.tests.test_loop.TestLoop, "
            "serial this txn started with 0x0101010101010101 1931-06-10 12:49:00.235294, "
            "serial currently committed 0x0202020202020202 1962-11-20 01:38:00.470588)"
        ))
